/* FSLinks.cpp
Original by Mikael Nordell, see below
Modifications (C) 2002 Niall Douglas under the GNU General Public Licence

Basically as was with removal of objects
*/
// Copyright 2000 Mikael Nordell (tamlin@algonet.se)
// Distributed without warranties. Use as you see fit, except for
// commercial purposes.
#include "FSLinks.h"
#include "accctrl.h"
#include "aclapi.h"
#include <assert.h>

// These have the wrong values in pre-W2k SDKs, why I redefine them here.
#if !defined(FSCTL_SET_REPARSE_POINT) || \
	(FSCTL_SET_REPARSE_POINT != 0x900a4)
#undef FSCTL_SET_REPARSE_POINT
#define FSCTL_SET_REPARSE_POINT  CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 41, METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif

#if !defined(FSCTL_DELETE_REPARSE_POINT) || \
	(FSCTL_DELETE_REPARSE_POINT != 0x900ac)
#undef FSCTL_DELETE_REPARSE_POINT
#define FSCTL_DELETE_REPARSE_POINT      CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 43, METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif

// Since MS apparently removed this struct (and its documentation) from
// the W2k SDK, but still refer to it in 'winioctl.h' for the specific
// IOCTLs, I decided to rename it and make it available.
// I've made some modifications to this one for easier access.
//
// Structure for FSCTL_SET_REPARSE_POINT, FSCTL_GET_REPARSE_POINT, and
// FSCTL_DELETE_REPARSE_POINT.
// This version of the reparse data buffer is only for Microsoft tags.

extern "C" struct TMN_REPARSE_DATA_BUFFER
{
    DWORD  ReparseTag;
    WORD   ReparseDataLength;
    WORD   Reserved;

	// IO_REPARSE_TAG_MOUNT_POINT specifics follow
    WORD   SubstituteNameOffset;
    WORD   SubstituteNameLength;
    WORD   PrintNameOffset;
    WORD   PrintNameLength;
    WCHAR  PathBuffer[1];
};

#define TMN_REPARSE_DATA_BUFFER_HEADER_SIZE \
			FIELD_OFFSET(TMN_REPARSE_DATA_BUFFER, SubstituteNameOffset)


BOOL CreateJunctionPoint(LPCTSTR szMountDir, LPCTSTR szDestDirArg)
{
	if (!szMountDir || !szDestDirArg || !szMountDir[0] || !szDestDirArg[0]) {
		return false;
	}
	// Ensure dest can handle reparse points
	{
		TCHAR szVol[4] = TEXT("x:\\");
		szVol[0] = szMountDir[0];
		DWORD dwCompLen;
		DWORD dwFsFlags;
		if (!GetVolumeInformation(szVol, 0, 0, 0, &dwCompLen, &dwFsFlags, 0, 0)) return false;

		if (!(dwFsFlags & FILE_SUPPORTS_REPARSE_POINTS)) {
			SetLastError(ERROR_REPARSE_ATTRIBUTE_CONFLICT);
			return false;
		}
	}

	TCHAR szDestDir[1024];
	if (szDestDirArg[0] == '\\' && szDestDirArg[1] == '?') {
		lstrcpy(szDestDir, szDestDirArg);
	} else {
		lstrcpy(szDestDir, TEXT("\\??\\"));
		TCHAR szFullDir[1024];
		LPTSTR pFilePart;
		if (!GetFullPathName(szDestDirArg, 1024, szFullDir, &pFilePart) ||
			GetFileAttributes(szFullDir) == -1)
		{
			return false;
		}
		lstrcat(szDestDir, szFullDir);
	}

	char szBuff[MAXIMUM_REPARSE_DATA_BUFFER_SIZE] = { 0 };
	TMN_REPARSE_DATA_BUFFER *rdb = (TMN_REPARSE_DATA_BUFFER*)szBuff;

	// Create the directory to mount with
	if(!CreateDirectory(szMountDir, NULL))
	{
		if(0xb7/*already exists*/!=GetLastError()) return 0;
	}
	// Prepare a suitable reparse data structure
	{
		const size_t nDestMountPointBytes = lstrlenW(szDestDir) * 2;

		rdb->ReparseTag           = IO_REPARSE_TAG_MOUNT_POINT;
		rdb->ReparseDataLength    = nDestMountPointBytes + 12;
		rdb->Reserved             = 0;
		rdb->SubstituteNameOffset = 0;
		rdb->SubstituteNameLength = nDestMountPointBytes;
		rdb->PrintNameOffset      = nDestMountPointBytes + 2;
		rdb->PrintNameLength      = 0;
		lstrcpyW(rdb->PathBuffer, szDestDir);
	}
	// Get a handle to the directory
	HANDLE h=0;
	h=CreateFile(szMountDir,
				GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT,
				0);
	if(INVALID_HANDLE_VALUE==h) goto handleerror;
	{
		USHORT par=COMPRESSION_FORMAT_DEFAULT;
		DWORD dwBytes;
		// Mark as blue
		if(!DeviceIoControl(h,
			FSCTL_SET_COMPRESSION,
			(LPVOID) &par,
			sizeof(par),
			NULL,
			0,
			&dwBytes,
			NULL)) goto handleerror;
		// Set reparse info
		if(!DeviceIoControl(h,
					FSCTL_SET_REPARSE_POINT,
					(LPVOID)rdb,
					rdb->ReparseDataLength + TMN_REPARSE_DATA_BUFFER_HEADER_SIZE,
					NULL,
					0,
					&dwBytes,
					0)) goto handleerror;
		if(!CloseHandle(h)) goto handleerror;
		h=0;
		// Mark as system (you get a nice warning not to delete)
		if(!SetFileAttributes(szMountDir, FILE_ATTRIBUTE_NOT_CONTENT_INDEXED|FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_SYSTEM)) goto handleerror;
		// To doubly prevent the user deleting, change its ACL's
#if 0
		// Turns out marking it no-delete doesn't help one jot
		{
			PACL dacl=NULL, newdacl=NULL;
			PSECURITY_DESCRIPTOR sd=NULL;
			__try
			{
				EXPLICIT_ACCESS ea={0};
				DWORD foo;
				if(ERROR_SUCCESS!=GetNamedSecurityInfo((LPTSTR) szMountDir, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION,
					NULL, NULL, &dacl, NULL, &sd)) goto handleerror;
				BuildExplicitAccessWithName(&ea, L"EVERYONE", DELETE, DENY_ACCESS,
					NO_INHERITANCE);
				/*
				ea.grfAccessPermissions=DELETE;
				ea.grfAccessMode=DENY_ACCESS;
				ea.grfInheritance=NO_INHERITANCE;
				ea.Trustee.pMultipleTrustee=NULL;
				ea.Trustee.MultipleTrusteeOperation=NO_MULTIPLE_TRUSTEE;
				ea.Trustee.TrusteeForm=TRUSTEE_IS_NAME;
				ea.Trustee.TrusteeType=TRUSTEE_IS_WELL_KNOWN_GROUP;
				ea.Trustee.ptstrName=L"EVERYONE";
				*/
				if(ERROR_SUCCESS!=SetEntriesInAcl(1, &ea, dacl, &newdacl)) goto handleerror;
				if(ERROR_SUCCESS!=SetNamedSecurityInfo((LPTSTR) szMountDir, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION,
					NULL, NULL, newdacl, NULL)) goto handleerror;
			}
			__finally
			{
				if(newdacl) LocalFree(newdacl);
				if(sd) LocalFree(sd);
			}
		}
#endif
	}
	if(0)
	{
handleerror:
		DWORD olderr=GetLastError();
		if(h) CloseHandle(h);
		RemoveDirectory(szMountDir);
		SetLastError(olderr);
		return false;
	}
	return true;
}

BOOL DeleteJunctionPoint(LPCTSTR szDir)
{
	REPARSE_GUID_DATA_BUFFER rgdb = { 0 };
	rgdb.ReparseTag = IO_REPARSE_TAG_MOUNT_POINT;
	DWORD dwBytes;
	HANDLE h;
	if(!SetFileAttributes(szDir, FILE_ATTRIBUTE_NORMAL)) return false;
	h=CreateFile(szDir,
				GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT,
				0);
	if(INVALID_HANDLE_VALUE==h) return false;
	if(!DeviceIoControl(h,
						FSCTL_DELETE_REPARSE_POINT,
						&rgdb,
						REPARSE_GUID_DATA_BUFFER_HEADER_SIZE,
						NULL,
						0,
						&dwBytes,
						0))
	{
		CloseHandle(h);
		return false;
	}
	if(!CloseHandle(h)) return false;
	return RemoveDirectory(szDir);
}

#if 0
DWORD GetJunctionPointInfo(	LPCTSTR szMountDir,
							LPTSTR  szDestBuff,
							DWORD   dwBuffSize)
{
	const DWORD dwFA = GetFileAttributes(szMountDir);
	if (dwFA == 0xffffffff || !(dwFA & FILE_ATTRIBUTE_REPARSE_POINT)) {
		SetLastError(ERROR_PATH_NOT_FOUND);
		return 0;
	}

	Reparse_Dir_HANDLE hDir(szMountDir);

	if (!hDir.IsValid()) {
		SetLastError(ERROR_PATH_NOT_FOUND);
		return 0;
	}

	char szBuff[MAXIMUM_REPARSE_DATA_BUFFER_SIZE];
	TMN_REPARSE_DATA_BUFFER& rdb = *(TMN_REPARSE_DATA_BUFFER*)szBuff;

	DWORD dwBytesReturned;
	if (!DeviceIoControl(hDir,
						FSCTL_GET_REPARSE_POINT,
						NULL,
						0,
						(LPVOID)&rdb,
						MAXIMUM_REPARSE_DATA_BUFFER_SIZE,
						&dwBytesReturned,
						0) ||
		!IsReparseTagValid(rdb.ReparseTag))
	{
		return 0;
	}

	if (dwBuffSize < rdb.SubstituteNameLength / sizeof(TCHAR) + sizeof(TCHAR))
	{
		return rdb.SubstituteNameLength / sizeof(TCHAR) + sizeof(TCHAR);
	}

#ifdef UNICODE
	lstrcpy(szDestBuff, rdb.PathBuffer);
#else
	if (!WideCharToMultiByte(CP_THREAD_ACP,
							0,
							rdb.PathBuffer,
							rdb.SubstituteNameLength / sizeof(WCHAR) + 1,
							szDestBuff,
							dwBuffSize,
							"",
							FALSE))
	{
		printf("WideCharToMultiByte failed (%d)\n", GetLastError());
		return 0;
	}
#endif
	return rdb.SubstituteNameLength / sizeof(TCHAR);
}

#endif